/**
 *
 */
package lazyj;

import lazyj.cache.GenericLastValuesCache;

/**
 * @author costing
 * @since May 14, 2018
 */
public class DBProperties extends ExtProperties {

	/**
	 * Database connection to execute the queries in
	 */
	final DBFunctions connection;

	/**
	 * Table name
	 */
	final String tableName;

	/**
	 * Primary key where the user-specified keys will be looked up by
	 */
	final String keyColumnName;

	/**
	 * Column holding the values
	 */
	final String valueColumnName;

	/**
	 * Last time the configuration was refreshed
	 */
	private long lastRefreshed = System.currentTimeMillis();

	/**
	 * If there is another set of keys to fall back to when the database doen't provide a value
	 */
	private ExtProperties parentConfig = null;

	/**
	 * Load the values from the specified database. By default the table is assumed to be called "config" with a "configkey" primary key and a "configvalue" column to actually hold the values
	 *
	 * @param connection
	 */
	public DBProperties(final DBFunctions connection) {
		this(connection, "config", "configkey", "configvalue", null);
	}

	/**
	 * Load the values from the specified database. By default the table is assumed to be called "config" with a "configkey" primary key and a "configvalue" column to actually hold the values
	 *
	 * @param connection
	 * @param parentConfig
	 *            configuration set to fall back to when the database query doesn't return anything for the given key
	 */
	public DBProperties(final DBFunctions connection, final ExtProperties parentConfig) {
		this(connection, "config", "configkey", "configvalue", parentConfig);
	}

	/**
	 * Load the values from the specified database, table name, primary key column and value column.
	 *
	 * @param connection
	 * @param tableName
	 * @param keyColumnName
	 * @param valueColumnName
	 * @param parentConfig
	 *            configuration set to fall back to when the database query doesn't return anything for the given key
	 */
	public DBProperties(final DBFunctions connection, final String tableName, final String keyColumnName, final String valueColumnName, final ExtProperties parentConfig) {
		this.connection = connection;
		this.tableName = tableName;
		this.keyColumnName = keyColumnName;
		this.valueColumnName = valueColumnName;
		this.parentConfig = parentConfig;
	}

	/**
	 * Cache the result of queries
	 */
	private final GenericLastValuesCache<String, String> dbCache = new GenericLastValuesCache<String, String>() {
		/**
		 *
		 */
		private static final long serialVersionUID = 4045527046095436303L;

		@Override
		protected synchronized String resolve(final String key) {
			if (connection == null)
				return null;

			if (!connection.query("SELECT " + valueColumnName + " FROM " + tableName + " WHERE " + keyColumnName + "=?", false, key))
				return null;

			if (!connection.moveNext())
				return null;

			return connection.gets(1, null);
		}
	};

	/*
	 * (non-Javadoc)
	 *
	 * @see lazyj.ExtProperties#gets(java.lang.String, java.lang.String, boolean)
	 */
	@Override
	public String gets(final String sKey, final String sDefault, final boolean bProcessQueries) {
		// reload config every minute
		if (System.currentTimeMillis() > lastRefreshed + 1000 * 60) {
			dbCache.refresh();
			lastRefreshed = System.currentTimeMillis();
		}

		final String value = dbCache.get(sKey);

		if (value != null)
			return value;

		if (parentConfig != null)
			return parentConfig.gets(sKey, sDefault, bProcessQueries);

		return sDefault;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see lazyj.ExtProperties#set(java.lang.String, java.lang.String)
	 */
	@Override
	public void set(String sKey, String sValue) {
		if (this.isReadOnly() || connection == null)
			throw new IllegalArgumentException("DB backed connection is read only, cannot set anything on it");

		if (!connection.query("UPDATE " + tableName + " SET " + valueColumnName + "=? WHERE " + keyColumnName + "=?;", false, sValue, sKey))
			return;

		if (connection.getUpdateCount() == 0)
			connection.query("INSERT INTO " + tableName + " (" + keyColumnName + "," + valueColumnName + ") VALUES (?,?);", false, sKey, sValue);

		// force the cache to remove this entry so that if needed it will be reloaded from the authoritative source
		dbCache.remove(sKey);
	}
}
