package lazyj;

/**
 * Wrap a PostgreSQL int8range column type with some simple operations. DBFunctions is now aware of this and can use it in {@link DBFunctions#composeInsert(String, java.util.Map)}
 * and {@link DBFunctions#composeUpdate(String, java.util.Map, java.util.Collection)}. If you want to parse a column simply say: <code>
 * EpochTimeRange range = new EpochTimeRange(db.gets("validity"));
 * </code>
 * 
 * @author costing
 * @since 2020-03-03
 */
public class EpochTimeRange {

	/**
	 * Whether or not the start of the interval is included in the range
	 */
	public boolean includingStart = true;

	/**
	 * Whether or not the end of the interval is included in the range
	 */
	public boolean includingEnd = false;

	/**
	 * Start time of the interval
	 */
	public long startTime = System.currentTimeMillis();

	/**
	 * End time of the interval
	 */
	public long endTime = Long.MAX_VALUE;

	/**
	 * default object is valid from now (inclusive) to infinity (exclusive :) )
	 */
	public EpochTimeRange() {

	}

	/**
	 * Initialize the interval endpoints from the database value (PostgreSQL text serialization of it)
	 * 
	 * @param dbValue DB column value
	 */
	public EpochTimeRange(final String dbValue) {
		includingStart = dbValue.charAt(0) == '[';
		includingEnd = dbValue.charAt(dbValue.length() - 1) == ']';

		int idx = dbValue.indexOf(',');

		startTime = Long.parseLong(dbValue.substring(1, idx));

		endTime = Long.parseLong(dbValue.substring(idx + 1, dbValue.length() - 1));
	}

	@Override
	public String toString() {
		return (includingStart ? "[" : "(") + startTime + "," + endTime + (includingEnd ? "]" : ")");
	}

	/**
	 * @return SQL representation of the value for this tuple
	 */
	public String toSQLString() {
		return "'" + toString() + "'::int8range";
	}
}
