/**
 *
 */
package lazyj;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

/**
 * @author costing
 * @since May 14, 2018
 */
public class FallbackProperties extends ExtProperties {
	/**
	 * Ordered list of configuration sets to query for the given key
	 */
	private final List<ExtProperties> configList = new ArrayList<>();

	/**
	 * Query the given list of configuration sets in order, returning the first defined value in one of them
	 *
	 * @param configProperties
	 */
	public FallbackProperties(final ExtProperties... configProperties) {
		if (configProperties != null)
			for (final ExtProperties config : configProperties)
				configList.add(config);
	}

	/**
	 * A different way of passing the ordered list of configuration sets to look into
	 *
	 * @param configProperties
	 */
	public FallbackProperties(final List<ExtProperties> configProperties) {
		if (configProperties != null)
			configList.addAll(configProperties);
	}

	/**
	 * Append a provider to the fallback list (having thus the lowest priority)
	 *
	 * @param config
	 */
	public void addProvider(final ExtProperties config) {
		addProvider(config, false);
	}

	/**
	 * Add another provider to the fallback list, either as the first element (with the highest priority) or as the last one (lowest priority).
	 *
	 * @param config
	 * @param first
	 *            <code>true</code> for a high priority provider, <code>false</code> for a low priority one
	 */
	public void addProvider(final ExtProperties config, final boolean first) {
		if (first)
			configList.add(0, config);
		else
			configList.add(config);
	}

	/**
	 * Remove a provider from the fallback list
	 *
	 * @param config
	 */
	public void removeProvider(final ExtProperties config) {
		configList.remove(config);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see lazyj.ExtProperties#gets(java.lang.String, java.lang.String, boolean)
	 */
	@Override
	public String gets(final String sKey, final String sDefault, final boolean bProcessQueries) {
		final Iterator<ExtProperties> itConfig = configList.iterator();

		while (itConfig.hasNext()) {
			final ExtProperties config = itConfig.next();

			final String value = itConfig.hasNext() ? config.gets(sKey, null, bProcessQueries) : config.gets(sKey, sDefault, bProcessQueries);

			if (value != null)
				return value;
		}

		return sDefault;
	}

	@Override
	public void set(final String sKey, final String sValue) {
		for (final ExtProperties config : configList)
			if (!config.isReadOnly())
				try {
					config.set(sKey, sValue);
					return;
				} catch (@SuppressWarnings("unused") final IllegalArgumentException e) {
					// just try the next one
				}

		throw new IllegalArgumentException("None of the " + configList.size() + " underlying configuration providers accepted the change");
	}

	@Override
	public boolean isReadOnly() {
		for (final ExtProperties config : configList)
			if (!config.isReadOnly())
				return false;

		return true;
	}

	@Override
	public Properties getProperties() {
		final Iterator<ExtProperties> itConfig = configList.iterator();

		final Properties ret = new Properties();

		while (itConfig.hasNext()) {
			final ExtProperties config = itConfig.next();

			final Properties prop = config.getProperties();

			for (final Object key : prop.keySet())
				ret.putIfAbsent(key, prop.get(key));
		}

		return ret;
	}
}
